{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE PackageImports #-}
{-# LANGUAGE TypeFamilies #-}

{-# LANGUAGE NoImplicitPrelude #-}

module Datarekisteri.Core.Types
    ( Base64
    , Date(..)
    , Email
    , KeyID(..)
    , Permission(..)
    , Permissions(..)
    , PhoneNumber
    , Scope(..)
    , Time(..)
    , TokenID(..)
    , Unit(Unit)
    , UserID(..)
    , addTime
    , base64Decode
    , base64Encode
    , hasPermission
    , renderDate
    , renderEmail
    , renderPhoneNumber
    , renderTime
    , toDate
    , toEmail
    , toPhoneNumber
    , toTime
    ) where

import Relude

import qualified "base64" Data.ByteString.Base64 as B64

import Data.Aeson (Value, ToJSON(..), FromJSON(..), Result(..), ToJSONKey(..), FromJSONKey(..), FromJSONKeyFunction(..), ToJSONKeyFunction(..), fromJSON, eitherDecodeStrict)
import Data.Aeson.Encoding (string)
import Data.Char (isSpace)
import qualified Data.Map as Map
import Data.Morpheus.Server.Types (SCALAR, TYPE)
import Data.Morpheus.Types (GQLType, DecodeScalar(..), KIND, EncodeScalar(..), ScalarValue(..))
import Data.Morpheus.Types.GQLScalar (scalarToJSON, scalarFromJSON)
import Data.Time (UTCTime, NominalDiffTime, addUTCTime, Day)
import Data.Time.Format.ISO8601 (iso8601Show, iso8601ParseM)
import Database.Persist.Class (PersistField(..))
import Database.Persist.PersistValue (PersistValue(..))
import Database.Persist.Postgresql.JSON () -- persistent instances for Aeson's Value
import Database.Persist.Sql (PersistFieldSql(..))
import Text.Email.Validate (EmailAddress, toByteString, validate, emailAddress)

import qualified Data.Map as Map
import qualified Data.Text as T

base64Encode :: ByteString -> Base64
base64Encode = Base64 . B64.encodeBase64

base64Decode :: Base64 -> Maybe ByteString
base64Decode (Base64 x) = either (const Nothing) Just $ B64.decodeBase64 $ encodeUtf8 x

toEmail :: Text -> Maybe Email
toEmail = fmap Email . emailAddress . encodeUtf8

renderEmail :: Email -> Text
renderEmail (Email x) = decodeUtf8 $ toByteString x

renderTime :: Time -> Text
renderTime (Time x) = toText $ iso8601Show x

toTime :: Text -> Maybe Time
toTime = fmap Time . iso8601ParseM . toString

toDate :: Text -> Maybe Date
toDate = fmap Date . iso8601ParseM . toString

renderDate :: Date -> Text
renderDate (Date x) = toText $ iso8601Show x

addTime :: NominalDiffTime -> Time -> Time
addTime diff (Time time) = Time $ addUTCTime diff time

data Scope = Profile UserID
           | Tokens UserID
           | Members
           | Applications
           deriving (Show, Eq, Ord, Read, Generic)

instance DecodeScalar Scope where
    decodeScalar (String s) = maybe (Left $ "invalid Scope: \"" <> s <> "\"") Right $ readMaybe (toString s)
    decodeScalar _ = Left "Invalid type for Scope, should be string"

instance EncodeScalar Scope where
    encodeScalar = String . show

instance GQLType Scope where type KIND Scope = SCALAR
instance ToJSON Scope where toJSON = scalarToJSON
instance FromJSON Scope where parseJSON = scalarFromJSON <=< parseJSON
instance ToJSONKey Scope where toJSONKey = ToJSONKeyText (fromString . show) (string . show)
instance FromJSONKey Scope
    where fromJSONKey = FromJSONKeyTextParser $
              either fail pure . (eitherDecodeStrict . encodeUtf8)

data Permission = None
                | ReadOnly
                | ReadWrite
                deriving (Show, Eq, Ord, Read, Generic)

instance DecodeScalar Permission where
    decodeScalar (String s) = maybe (Left $ "invalid Permission: \"" <> s <> "\"") Right $ readMaybe (toString s)
    decodeScalar _ = Left "Invalid type for Permission, should be string"

instance EncodeScalar Permission where
    encodeScalar = String . show

instance GQLType Permission where type KIND Permission = SCALAR
instance ToJSON Permission where toJSON = scalarToJSON
instance FromJSON Permission where parseJSON = scalarFromJSON <=< parseJSON

hasPermission :: Scope -> Permission -> Permissions -> Bool
hasPermission scope permission (Permissions permissionMap) =
    Map.findWithDefault None scope permissionMap >= permission

newtype Permissions = Permissions (Map Scope Permission)
    deriving (Show, Eq, Read, Generic, ToJSON, FromJSON, Semigroup, Monoid)

instance GQLType Permissions where type KIND Permissions = TYPE

instance PersistField Permissions where
    toPersistValue = toPersistValue . toJSON
    fromPersistValue = resultToEither . fromJSON <=< fromPersistValue
        where resultToEither (Success x) = Right x
              resultToEither (Error err) = Left $ T.pack err

instance PersistFieldSql Permissions where sqlType _ = sqlType (Proxy :: Proxy Value)

instance Ord Permissions where
    (Permissions a) <= (Permissions b) =
        and $ Map.intersectionWith (<=) a $ b <> (const None <$> Map.difference a b)

newtype PhoneNumber = PhoneNumber Text deriving (Show, Generic)

renderPhoneNumber :: PhoneNumber -> Text
renderPhoneNumber (PhoneNumber txt) = txt

toPhoneNumber :: Text -> Maybe PhoneNumber
toPhoneNumber txt = do
    guard $ not $ T.null txt
    guard $ T.head txt `elem` ['+','0']
    guard $ T.all (`elem` (" 0123456789" :: [Char])) $ T.tail txt
    pure $ PhoneNumber txt

instance Eq PhoneNumber where
    a == b = normalize a == normalize b
        where normalize (PhoneNumber txt) = case T.uncons txt of
                                            Just ('0', rest) -> "+358" <> T.filter (not . isSpace) rest
                                            Just x -> T.filter (not . isSpace) (uncurry T.cons x)
                                            Nothing -> error "PhoneNumber must not be empty"

instance DecodeScalar PhoneNumber where
    decodeScalar (String s) =
        maybe (Left $ "Couldn't parse \"" <> s <> "\" as a PhoneNumber") Right $ toPhoneNumber s
    decodeScalar _ = Left "Invalid type for PhoneNumber, should be string"

instance EncodeScalar PhoneNumber where
    encodeScalar = String . renderPhoneNumber

instance GQLType PhoneNumber where type KIND PhoneNumber = SCALAR
instance ToJSON PhoneNumber where toJSON = scalarToJSON
instance FromJSON PhoneNumber where parseJSON = scalarFromJSON <=< parseJSON

newtype UserID = UserID Integer deriving (Eq, Show, Generic, Ord, Read)

instance DecodeScalar UserID where
    decodeScalar (String s) = first (const $ "invalid UserID: \"" <> s <> "\"") $
        UserID <$> readEither (toString s)
    decodeScalar _ = Left "Invalid type for UserID, should be string"

instance EncodeScalar UserID where
    encodeScalar (UserID x) = String $ show x

instance GQLType UserID where type KIND UserID = SCALAR
instance ToJSON UserID where toJSON = scalarToJSON
instance FromJSON UserID where parseJSON = scalarFromJSON <=< parseJSON

newtype KeyID = KeyID Integer deriving (Eq, Show, Generic)

instance DecodeScalar KeyID where
    decodeScalar (String s) = first (const $ "invalid KeyID: \"" <> s <> "\"") $
        KeyID <$> readEither (toString s)
    decodeScalar _ = Left "Invalid type for KeyID, should be string"

instance EncodeScalar KeyID where
    encodeScalar (KeyID s) = String $ show s

instance GQLType KeyID where type KIND KeyID = SCALAR
instance ToJSON KeyID where toJSON = scalarToJSON
instance FromJSON KeyID where parseJSON = scalarFromJSON <=< parseJSON

newtype TokenID = TokenID Integer deriving (Eq, Show, Generic)

instance DecodeScalar TokenID where
    decodeScalar (String s) = first (const $ "invalid TokenID: \"" <> s <> "\"") $
        TokenID <$> readEither (toString s)
    decodeScalar _ = Left "Invalid type for TokenID, should be string"

instance EncodeScalar TokenID where
    encodeScalar (TokenID s) = String $ show s

instance GQLType TokenID where type KIND TokenID = SCALAR
instance ToJSON TokenID where toJSON = scalarToJSON
instance FromJSON TokenID where parseJSON = scalarFromJSON <=< parseJSON

newtype Base64 = Base64 Text
    deriving (Eq, Show, ToJSON, FromJSON, Generic, DecodeScalar, EncodeScalar)

instance GQLType Base64 where type KIND Base64 = SCALAR

newtype Email = Email EmailAddress deriving (Eq, Show)

instance DecodeScalar Email where
    decodeScalar (String s) = maybe (Left $ "Couldn't parse \"" <> s <> "\" as an email address") (Right . Email) $ emailAddress $ encodeUtf8 s
    decodeScalar _ = Left "Invalid type for Email, should be string"

instance EncodeScalar Email where
    encodeScalar = String . renderEmail

instance GQLType Email where type KIND Email = SCALAR
instance ToJSON Email where toJSON = scalarToJSON
instance FromJSON Email where parseJSON = scalarFromJSON <=< parseJSON 

instance PersistField Email where
    toPersistValue = PersistText . renderEmail
    fromPersistValue (PersistText email) =
        second Email $ first toText $ validate $ encodeUtf8 email
    fromPersistValue x = Left $ "Wrong type for Email: " <> show x

newtype Time = Time UTCTime deriving (Eq, Show, PersistField, PersistFieldSql)

instance DecodeScalar Time where
    decodeScalar (String s) = maybe
        (Left $ "Couldn't parse \"" <> s <> "\" as an ISO8601 timestamp")
        (Right) $ toTime s
    decodeScalar _ = Left "Invalid type for Time, should be string"

instance EncodeScalar Time where
    encodeScalar  = String . renderTime

instance GQLType Time where type KIND Time = SCALAR
instance ToJSON Time where toJSON = scalarToJSON
instance FromJSON Time where parseJSON = scalarFromJSON <=< parseJSON

newtype Date = Date Day deriving (Eq, Show, PersistField, Generic)

instance DecodeScalar Date where
    decodeScalar (String s) = maybe
        (Left $ "Couldn't parse \"" <> s <> "\" as an ISO8601 date")
        (Right) $ toDate s
    decodeScalar _ = Left "Invalid type for Date, should be string"

instance EncodeScalar Date where
    encodeScalar = String . renderDate

instance GQLType Date where type KIND Date = SCALAR
instance ToJSON Date where toJSON = scalarToJSON
instance FromJSON Date where parseJSON = scalarFromJSON <=< parseJSON

data Unit = Unit | Unit2 deriving (Eq, Generic, Show) -- XXX: why not just Unit

instance GQLType Unit
instance ToJSON Unit
instance FromJSON Unit
